@@ -1,6 +1,6 @@
 #!/bin/sh
 
-# Copyright (c) 2012 - 2013 dak180 and contributors. See
+# Copyright (c) 2012 - 2014 dak180 and contributors. See
 # http://opensource.org/licenses/mit-license.php or the included
 # COPYING.md for licence terms.
 #
@@ -10,7 +10,7 @@
 # Usage message.
 arUsage() {
 	cat > "/dev/stderr" << EOF
-usage: ./autorevision {-t output-type | -s symbol} [-o cache-file [-f] ] [-V]
+usage: autorevision {-t output-type | -s symbol} [-o cache-file [-f] ] [-V]
 	Options include:
 	-t output-type		= specify output type
 	-s symbol		= specify symbol output
@@ -20,9 +20,10 @@ usage: ./autorevision {-t output-type | -s symbol} [-o cache-file [-f] ] [-V]
 	-V			= emit version and exit
 	-?			= help message
 
-The folowing are valid output types:
+The following are valid output types:
 	h			= Header for use with c/c++
 	xcode			= Header useful for populating info.plist files
+	swift			= Swift file
 	sh			= Bash sytax
 	py			= Python file
 	pl			= Perl file
@@ -39,11 +40,13 @@ The folowing are valid output types:
 The following are valid symbols:
 	VCS_TYPE
 	VCS_BASENAME
+	VCS_UUID
 	VCS_NUM
 	VCS_DATE
 	VCS_BRANCH
 	VCS_TAG
 	VCS_TICK
+	VCS_EXTRA
 	VCS_FULL_HASH
 	VCS_SHORT_HASH
 	VCS_WC_MODIFIED
@@ -92,6 +95,10 @@ elif [ -z "${VAROUT}" ] && [ -z "${AFILETYPE}" ]; then
 elif [ -z "${CACHEFILE}" ] && [ "${CACHEFORCE}" = "1" ]; then
 	# If -f is specified without -o:
 	arUsage
+elif [ ! -f "${CACHEFILE}" ] && [ "${CACHEFORCE}" = "1" ]; then
+	# If we are forced to use the cache but it does not exist.
+	echo "error: Cache forced but no cache found." 1>&2
+	exit 1
 fi
 
 # Make sure that the path we are given is one we can source
@@ -110,6 +117,11 @@ gitRepo() {
 
 	VCS_BASENAME="$(basename "${PWD}")"
 
+	VCS_UUID="$(git rev-list --max-parents=0 --date-order --reverse HEAD 2>/dev/null | sed -n 1p)"
+	if [ -z "${VCS_UUID}" ]; then
+		VCS_UUID="$(git rev-list --topo-order HEAD | tail -n 1)"
+	fi
+
 	 # Is the working copy clean?
 	test -z "$(git status --untracked-files=normal --porcelain)"
 	VCS_WC_MODIFIED="${?}"
@@ -163,6 +175,8 @@ hgRepo() {
 
 	VCS_BASENAME="$(basename "${PWD}")"
 
+	VCS_UUID="$(hg log -r "0" -l 1 --template '{node}\n')"
+
 	# Is the working copy clean?
 	test -z "$(hg status -duram)"
 	VCS_WC_MODIFIED="${?}"
@@ -206,6 +220,9 @@ bzrRepo() {
 
 	VCS_BASENAME="$(basename "${PWD}")"
 
+	# Currently unimplemented because more investigation is needed.
+	VCS_UUID=""
+
 	# Is the working copy clean?
 	bzr version-info --custom --template='{clean}\n' | grep -q '1'
 	VCS_WC_MODIFIED="${?}"
@@ -256,6 +273,8 @@ svnRepo() {
 	*) VCS_BASENAME="$(basename "${PWD}")" ;;
 	esac
 
+	VCS_UUID="$(svn info | sed -n -e 's:Repository UUID\: ::p')"
+
 	# Cache svnversion output
 	local SVNVERSION="$(svnversion)"
 
@@ -266,7 +285,7 @@ svnRepo() {
 			VCS_WC_MODIFIED="1"
 		;;
 		1)
-			if [ ! -z "${UNTRACKEDFILES}"]; then
+			if [ ! -z "${UNTRACKEDFILES}" ]; then
 			# `svnversion` does not detect untracked files and `svn status` is really slow, so only run it if we really have to.
 				if [ -z "$(svn status)" ]; then
 					VCS_WC_MODIFIED="0"
@@ -311,9 +330,9 @@ svnRepo() {
 	*) VCS_BRANCH="" ;;
 	esac
 
-	# Current or last tag ancestor (empty if no tags). But "current tag"
-	# can't be extracted reliably because Subversion doesn't have tags the
-	# way other VCSes do.
+	# Current or last tag ancestor (empty if no tags). But "current
+	# tag" can't be extracted reliably because Subversion doesn't
+	# have tags the way other VCSes do.
 	VCS_TAG=""
 	VCS_TICK=""
 
@@ -332,11 +351,13 @@ hOutput() {
 
 #define VCS_TYPE		"${VCS_TYPE}"
 #define VCS_BASENAME	"${VCS_BASENAME}"
+#define VCS_UUID		"${VCS_UUID}"
 #define VCS_NUM			${VCS_NUM}
 #define VCS_DATE		"${VCS_DATE}"
 #define VCS_BRANCH		"${VCS_BRANCH}"
 #define VCS_TAG			"${VCS_TAG}"
 #define VCS_TICK		${VCS_TICK}
+#define VCS_EXTRA		"${VCS_EXTRA}"
 
 #define VCS_FULL_HASH		"${VCS_FULL_HASH}"
 #define VCS_SHORT_HASH		"${VCS_SHORT_HASH}"
@@ -358,11 +379,13 @@ xcodeOutput() {
 
 #define VCS_TYPE		${VCS_TYPE}
 #define VCS_BASENAME	${VCS_BASENAME}
+#define VCS_UUID		${VCS_UUID}
 #define VCS_NUM			${VCS_NUM}
 #define VCS_DATE		${VCS_DATE}
 #define VCS_BRANCH		${VCS_BRANCH}
 #define VCS_TAG			${VCS_TAG}
 #define VCS_TICK		${VCS_TICK}
+#define VCS_EXTRA		${VCS_EXTRA}
 
 #define VCS_FULL_HASH		${VCS_FULL_HASH}
 #define VCS_SHORT_HASH		${VCS_SHORT_HASH}
@@ -375,6 +398,52 @@ xcodeOutput() {
 EOF
 }
 
+# For Swift output
+swiftOutput() {
+	case "${VCS_WC_MODIFIED}" in
+		0) VCS_WC_MODIFIED="false" ;;
+		1) VCS_WC_MODIFIED="true" ;;
+	esac
+	# For values that may not exist depending on the type of repo we
+	# have read from, set them to `nil` when they are empty.
+	if [ -z "${VCS_UUID}" ]; then
+		VCS_UUID="nil"
+	else
+		VCS_UUID="\"${VCS_UUID}\""
+	fi
+	if [ -z "${VCS_TAG}" ]; then
+		VCS_TAG="nil"
+	else
+		VCS_TAG="\"${VCS_TAG}\""
+	fi
+	: "${VCS_TICK:="nil"}"
+	if [ -z "${VCS_EXTRA}" ]; then
+		VCS_EXTRA="nil"
+	else
+		VCS_EXTRA="\"${VCS_EXTRA}\""
+	fi
+	cat > "${TARGETFILE}" << EOF
+/* Generated by autorevision - do not hand-hack! */
+
+let VCS_TYPE			= "${VCS_TYPE}"
+let VCS_BASENAME		= "${VCS_BASENAME}"
+let VCS_UUID:	String?	= ${VCS_UUID}
+let VCS_NUM:	Int		= ${VCS_NUM}
+let VCS_DATE			= "${VCS_DATE}"
+let VCS_BRANCH:	String	= "${VCS_BRANCH}"
+let VCS_TAG:	String?	= ${VCS_TAG}
+let VCS_TICK:	Int?	= ${VCS_TICK}
+let VCS_EXTRA:	String?	= ${VCS_EXTRA}
+
+let VCS_FULL_HASH:		String	= "${VCS_FULL_HASH}"
+let VCS_SHORT_HASH:		String	= "${VCS_SHORT_HASH}"
+
+let VCS_WC_MODIFIED:	Bool	= ${VCS_WC_MODIFIED}
+
+/* end */
+EOF
+}
+
 # For bash output
 shOutput() {
 	cat > "${TARGETFILE}" << EOF
@@ -382,16 +451,18 @@ shOutput() {
 
 VCS_TYPE="${VCS_TYPE}"
 VCS_BASENAME="${VCS_BASENAME}"
-VCS_NUM=${VCS_NUM}
+VCS_UUID="${VCS_UUID}"
+VCS_NUM="${VCS_NUM}"
 VCS_DATE="${VCS_DATE}"
 VCS_BRANCH="${VCS_BRANCH}"
 VCS_TAG="${VCS_TAG}"
-VCS_TICK=${VCS_TICK}
+VCS_TICK="${VCS_TICK}"
+VCS_EXTRA="${VCS_EXTRA}"
 
 VCS_FULL_HASH="${VCS_FULL_HASH}"
 VCS_SHORT_HASH="${VCS_SHORT_HASH}"
 
-VCS_WC_MODIFIED=${VCS_WC_MODIFIED}
+VCS_WC_MODIFIED="${VCS_WC_MODIFIED}"
 
 # end
 EOF
@@ -408,11 +479,13 @@ pyOutput() {
 
 VCS_TYPE = "${VCS_TYPE}"
 VCS_BASENAME = "${VCS_BASENAME}"
+VCS_UUID = "${VCS_UUID}"
 VCS_NUM = ${VCS_NUM}
 VCS_DATE = "${VCS_DATE}"
 VCS_BRANCH = "${VCS_BRANCH}"
 VCS_TAG = "${VCS_TAG}"
 VCS_TICK = ${VCS_TICK}
+VCS_EXTRA = "${VCS_EXTRA}"
 
 VCS_FULL_HASH = "${VCS_FULL_HASH}"
 VCS_SHORT_HASH = "${VCS_SHORT_HASH}"
@@ -430,11 +503,13 @@ plOutput() {
 
 \$VCS_TYPE = "${VCS_TYPE}";
 \$VCS_BASENAME = "${VCS_BASENAME}"
+\$VCS_UUID = "${VCS_UUID}"
 \$VCS_NUM = ${VCS_NUM};
 \$VCS_DATE = "${VCS_DATE}";
 \$VCS_BRANCH = "${VCS_BRANCH}";
 \$VCS_TAG = "${VCS_TAG}";
 \$VCS_TICK = ${VCS_TICK};
+\$VCS_EXTRA = "${VCS_EXTRA}";
 
 \$VCS_FULL_HASH = "${VCS_FULL_HASH}";
 \$VCS_SHORT_HASH = "${VCS_SHORT_HASH}";
@@ -456,11 +531,13 @@ luaOutput() {
 
 VCS_TYPE = "${VCS_TYPE}"
 VCS_BASENAME = "${VCS_BASENAME}"
+VCS_UUID = "${VCS_UUID}"
 VCS_NUM = ${VCS_NUM}
 VCS_DATE = "${VCS_DATE}"
 VCS_BRANCH = "${VCS_BRANCH}"
 VCS_TAG = "${VCS_TAG}"
 VCS_TICK = ${VCS_TICK}
+VCS_EXTRA = "${VCS_EXTRA}"
 
 VCS_FULL_HASH = "${VCS_FULL_HASH}"
 VCS_SHORT_HASH = "${VCS_SHORT_HASH}"
@@ -483,11 +560,13 @@ phpOutput() {
 
 $GLOBALS["VCS_TYPE"] = "${VCS_TYPE}";
 $GLOBALS["VCS_BASENAME"] = "${VCS_BASENAME}";
+$GLOBALS["VCS_UUID"] = "${VCS_UUID}";
 $GLOBALS["VCS_NUM"] = ${VCS_NUM};
 $GLOBALS["VCS_DATE"] = "${VCS_DATE}";
 $GLOBALS["VCS_BRANCH"] = "${VCS_BRANCH}";
 $GLOBALS["VCS_TAG"] = "${VCS_TAG}";
 $GLOBALS["VCS_TICK"] = ${VCS_TICK};
+$GLOBALS["VCS_EXTRA"] = "${VCS_EXTRA}";
 
 $GLOBALS["VCS_FULL_HASH"] = "${VCS_FULL_HASH}";
 $GLOBALS["VCS_SHORT_HASH"] = "${VCS_SHORT_HASH}";
@@ -510,11 +589,13 @@ iniOutput() {
 [VCS]
 VCS_TYPE = "${VCS_TYPE}"
 VCS_BASENAME = "${VCS_BASENAME}"
+VCS_UUID = "${VCS_UUID}"
 VCS_NUM = ${VCS_NUM}
 VCS_DATE = "${VCS_DATE}"
 VCS_BRANCH = "${VCS_BRANCH}"
 VCS_TAG = "${VCS_TAG}"
 VCS_TICK = ${VCS_TICK}
+VCS_EXTRA = "${VCS_EXTRA}"
 VCS_FULL_HASH = "${VCS_FULL_HASH}"
 VCS_SHORT_HASH = "${VCS_SHORT_HASH}"
 VCS_WC_MODIFIED = ${VCS_WC_MODIFIED}
@@ -534,11 +615,13 @@ jsOutput() {
 var autorevision = {
 	VCS_TYPE: "${VCS_TYPE}",
 	VCS_BASENAME: "${VCS_BASENAME}",
+	VCS_UUID: "${VCS_UUID}",
 	VCS_NUM: ${VCS_NUM},
 	VCS_DATE: "${VCS_DATE}",
 	VCS_BRANCH: "${VCS_BRANCH}",
 	VCS_TAG: "${VCS_TAG}",
 	VCS_TICK: ${VCS_TICK},
+	VCS_EXTRA: "${VCS_EXTRA}",
 
 	VCS_FULL_HASH: "${VCS_FULL_HASH}",
 	VCS_SHORT_HASH: "${VCS_SHORT_HASH}",
@@ -565,11 +648,13 @@ jsonOutput() {
 {
 	"VCS_TYPE": "${VCS_TYPE}",
 	"VCS_BASENAME": "${VCS_BASENAME}",
+	"VCS_UUID": "${VCS_UUID}",
 	"VCS_NUM": ${VCS_NUM},
 	"VCS_DATE": "${VCS_DATE}",
 	"VCS_BRANCH":"${VCS_BRANCH}",
 	"VCS_TAG": "${VCS_TAG}",
 	"VCS_TICK": ${VCS_TICK},
+	"VCS_EXTRA": "${VCS_EXTRA}",
 
 	"VCS_FULL_HASH": "${VCS_FULL_HASH}",
 	"VCS_SHORT_HASH": "${VCS_SHORT_HASH}",
@@ -588,16 +673,16 @@ javaOutput() {
 	cat > "${TARGETFILE}" << EOF
 /* Generated by autorevision - do not hand-hack! */
 
-import java.util.Date;
-
 public class autorevision {
     public static final String VCS_TYPE = "${VCS_TYPE}";
     public static final String VCS_BASENAME = "${VCS_BASENAME}";
+    public static final String VCS_UUID = "${VCS_UUID}";
     public static final long VCS_NUM = ${VCS_NUM};
     public static final String VCS_DATE = "${VCS_DATE}";
     public static final String VCS_BRANCH = "${VCS_BRANCH}";
     public static final String VCS_TAG = "${VCS_TAG}";
     public static final long VCS_TICK = ${VCS_TICK};
+    public static final String VCS_EXTRA = "${VCS_EXTRA}";
 
     public static final String VCS_FULL_HASH = "${VCS_FULL_HASH}";
     public static final String VCS_SHORT_HASH = "${VCS_SHORT_HASH}";
@@ -618,11 +703,13 @@ javapropOutput() {
 
 VCS_TYPE=${VCS_TYPE}
 VCS_BASENAME=${VCS_BASENAME}
+VCS_UUID=${VCS_UUID}
 VCS_NUM=${VCS_NUM}
 VCS_DATE=${VCS_DATE}
 VCS_BRANCH=${VCS_BRANCH}
 VCS_TAG=${VCS_TAG}
 VCS_TICK=${VCS_TICK}
+VCS_EXTRA=${VCS_EXTRA}
 
 VCS_FULL_HASH=${VCS_FULL_HASH}
 VCS_SHORT_HASH=${VCS_SHORT_HASH}
@@ -636,13 +723,15 @@ m4Output() {
 	cat > "${TARGETFILE}" << EOF
 define(\`VCS_TYPE', \`${VCS_TYPE}')dnl
 define(\`VCS_BASENAME', \`${VCS_BASENAME}')dnl
+define(\`VCS_UUID', \`${VCS_UUID}')dnl
 define(\`VCS_NUM', \`${VCS_NUM}')dnl
 define(\`VCS_DATE', \`${VCS_DATE}')dnl
 define(\`VCS_BRANCH', \`${VCS_BRANCH}')dnl
 define(\`VCS_TAG', \`${VCS_TAG}')dnl
 define(\`VCS_TICK', \`${VCS_TICK}')dnl
+define(\`VCS_EXTRA', \`${VCS_EXTRA}')dnl
 define(\`VCS_FULLHASH', \`${VCS_FULL_HASH}')dnl
-define(\`VCS_SHIRTHASH', \`${VCS_SHORT_HASH}')dnl
+define(\`VCS_SHORTHASH', \`${VCS_SHORT_HASH}')dnl
 define(\`VCS_WC_MODIFIED', \`${VCS_WC_MODIFIED}')dnl
 EOF
 }
@@ -657,11 +746,13 @@ texOutput() {
 % Generated by autorevision - do not hand-hack!
 \def \vcsType {${VCS_TYPE}}
 \def \vcsBasename {${VCS_BASENAME}}
+\def \vcsUUID {${VCS_UUID}}
 \def \vcsNum {${VCS_NUM}}
 \def \vcsDate {${VCS_DATE}}
 \def \vcsBranch {${VCS_BRANCH}}
 \def \vcsTag {${VCS_TAG}}
 \def \vcsTick {${VCS_TICK}}
+\def \vcsExtra {${VCS_EXTRA}}
 \def \vcsFullHash {${VCS_FULL_HASH}}
 \def \vcsShortHash {${VCS_SHORT_HASH}}
 \def \vcsWCModified {${VCS_WC_MODIFIED}}
@@ -671,26 +762,105 @@ EOF
 
 
 
-# Detect and collect repo data.
+# Helper functions
+# Count path segments
+pathSegment() {
+	local pathz="${1}"
+	local depth="0"
+
+	if [ ! -z "${pathz}" ]; then
+		while [ ! "${pathz}" = "/" ] && [ ! "${pathz}" = "." ]; do
+			pathz="$(dirname "${pathz}")"
+			depth="$((depth+1))"
+		done
+	fi
+	echo "${depth}"
+}
+
+# Largest of four numbers
+multiCompare() {
+	local larger="${1}"
+	local numA="${2}"
+	local numB="${3}"
+	local numC="${4}"
+
+	[ "${numA}" -gt "${larger}" ] && larger="${numA}"
+	[ "${numB}" -gt "${larger}" ] && larger="${numB}"
+	[ "${numC}" -gt "${larger}" ] && larger="${numC}"
+	echo "${larger}"
+}
+
+# Test for repositories
+repoTest() {
+	REPONUM="0"
+	if [ ! -z "$(git rev-parse HEAD 2>/dev/null)" ]; then
+		local gitPath="$(git rev-parse --show-toplevel)"
+		local gitDepth="$(pathSegment "${gitPath}")"
+		REPONUM="$((REPONUM+1))"
+	else
+		local gitDepth="0"
+	fi
+	if [ ! -z "$(hg root 2>/dev/null)" ]; then
+		local hgPath="$(hg root 2>/dev/null)"
+		local hgDepth="$(pathSegment "${hgPath}")"
+		REPONUM="$((REPONUM+1))"
+	else
+		local hgDepth="0"
+	fi
+	if [ ! -z "$(bzr root 2>/dev/null)" ]; then
+		local bzrPath="$(bzr root 2>/dev/null)"
+		local bzrDepth="$(pathSegment "${bzrPath}")"
+		REPONUM="$((REPONUM+1))"
+	else
+		local bzrDepth="0"
+	fi
+	if [ ! -z "$(svn info 2>/dev/null)" ]; then
+		local stringz="Working Copy Root Path: "
+		local svnPath="$(svn info | grep "${stringz}" | tr -d "${stringz}")"
+		local svnDepth="$(pathSegment "${svnPath}")"
+		REPONUM="$((REPONUM+1))"
+	else
+		local svnDepth="0"
+	fi
+
+	# Do not do more work then we have to.
+	if [ "${REPONUM}" = "0" ]; then
+		return
+	fi
+
+	# Figure out which repo is the deepest and use it.
+	local wonRepo="$(multiCompare "${gitDepth}" "${hgDepth}" "${bzrDepth}" "${svnDepth}")"
+	if [ "${wonRepo}" = "${gitDepth}" ]; then
+		gitRepo
+	elif [ "${wonRepo}" = "${hgDepth}" ]; then
+		hgRepo
+	elif [ "${wonRepo}" = "${bzrDepth}" ]; then
+		bzrRepo
+	elif [ "${wonRepo}" = "${svnDepth}" ]; then
+		svnRepo
+	fi
+}
+
+
+
+# Detect which repos we are in and gather data.
 if [ -f "${CACHEFILE}" ] && [ "${CACHEFORCE}" = "1" ]; then
 	# When requested only read from the cache to populate our symbols.
 	. "${CACHEFILE}"
-elif [ ! -z "$(git rev-parse HEAD 2>/dev/null)" ]; then
-	gitRepo
-elif [ ! -z "$(hg root 2>/dev/null)" ]; then
-	hgRepo
-elif [ ! -z "$(bzr root 2>/dev/null)" ]; then
-	bzrRepo
-elif [ ! -z "$(svn info 2>/dev/null)" ]; then
-	svnRepo
-elif [ -f "${CACHEFILE}" ]; then
-	# We are not in a repo; try to use a previously generated cache to populate our symbols.
-	. "${CACHEFILE}"
-	# Do not overwrite the cache if we know we are not going to write anything new.
-	CACHEFORCE="1"
 else
-	echo "error: No repo or cache detected." 1>&2
-	exit 1
+	# If a value is not set through the environment set VCS_EXTRA to nothing.
+	: "${VCS_EXTRA:=""}"
+	repoTest
+
+	if [ -f "${CACHEFILE}" ] && [ "${REPONUM}" = "0" ]; then
+		# We are not in a repo; try to use a previously generated cache to populate our symbols.
+		. "${CACHEFILE}"
+		# Do not overwrite the cache if we know we are not going to write anything new.
+		CACHEFORCE="1"
+	elif [ "${REPONUM}" = "0" ]; then
+		echo "error: No repo or cache detected." 1>&2
+		exit 1
+	fi
 fi
 
 
@@ -729,6 +899,8 @@ if [ ! -z "${AFILETYPE}" ]; then
 		hOutput
 	elif [ "${AFILETYPE}" = "xcode" ]; then
 		xcodeOutput
+	elif [ "${AFILETYPE}" = "swift" ]; then
+		swiftOutput
 	elif [ "${AFILETYPE}" = "sh" ]; then
 		shOutput
 	elif [ "${AFILETYPE}" = "py" ] || [ "${AFILETYPE}" = "python" ]; then
@@ -762,6 +934,15 @@ fi
 
 # If requested, make a cache file.
 if [ ! -z "${CACHEFILE}" ] && [ ! "${CACHEFORCE}" = "1" ]; then
-	TARGETFILE="${CACHEFILE}"
+	TARGETFILE="${CACHEFILE}.tmp"
 	shOutput
+
+	# Check to see if there have been any actual changes.
+	if [ ! -f "${CACHEFILE}" ]; then
+		mv -f "${CACHEFILE}.tmp" "${CACHEFILE}"
+	elif cmp -s "${CACHEFILE}.tmp" "${CACHEFILE}"; then
+		rm -f "${CACHEFILE}.tmp"
+	else
+		mv -f "${CACHEFILE}.tmp" "${CACHEFILE}"
+	fi
 fi
